from collections import defaultdict 
import time
import datetime
import platform
import os
import threading

from mute_control import MuteState
from led import LedState
import libamplifier_lib as amp
#import amp
from pylogger import bllogger
import functools
import traceback

def LockCFunction(func):
    @functools.wraps(func)
    def runFunc(*args, **kwargs):
        rtn = ""
        #print "function about to take config lock %s" % func
        config = amp.GetConfigLock() 
        #print "function took config lock %s" % func
        amplifier = amp.GetAmplifierSettingLock()
        try:
            rtn = func(*args, **kwargs)
        except Exception:
            tb = traceback.format_exc()
            bllogger('err', "Error encounterd in function (%r) " %func)
            bllogger('err', "%r" %tb)
        finally:
            amp.ReleaseAmplifierSettingLock(amplifier)
            #print "function about to release config lock %s" % func
            amp.ReleaseConfigLock(config)
            #print "function released config lock %s" % func
        return rtn
    return runFunc

class AmplifierSetting:

    def __init__(self, gpio_handle):
        self.gpio_handle = gpio_handle
        self.amp_setting = amp.GetAmplifierSetting()

        self.led_state = LedState()
        led_control = threading.Thread(target=self.led_state.led_thread)
        led_control.start()

        self.mute_state = MuteState(gpio_handle)
        mute_control = threading.Thread(target=self.mute_state.mute_thread)
        mute_control.start()

        self.room = self.amp_setting.GetRoom()

        self.lock = threading.RLock()

        self.update_available = False

        self.config = amp.SimpleConfig("fatal_config", "")
        self.config.Load("", False)

        self.room_table = {}
        self.fatal_shutdown = False
        self.thermal_time = 10   #in seconds

        self.find_id()
        self.power_time_2 = 0.262
        self.surge_time_1 = 1.835
        self.surge_time_2 = 0.262
        self.safe_time = 2.786

        self.error_dict = defaultdict(lambda:False)
        self.fatal_errors = [
                amp.key_map.LEFT_POS_15VOLT_ERROR,
                amp.key_map.RIGHT_POS_15VOLT_ERROR,
                amp.key_map.LEFT_NEG_15VOLT_ERROR,
                amp.key_map.RIGHT_NEG_15VOLT_ERROR,
                amp.key_map.LEFT_POS_IV_ERROR,
                amp.key_map.RIGHT_POS_IV_ERROR,
                amp.key_map.LEFT_NEG_IV_ERROR,
                amp.key_map.RIGHT_NEG_IV_ERROR,
                ]

        self.transient_errors = {
                amp.key_map.LEFT_INPUT_DC: '',
                amp.key_map.RIGHT_INPUT_DC: '',
                amp.key_map.LEFT_CLIPPING: '',
                amp.key_map.RIGHT_CLIPPING: '',
                amp.key_map.LEFT_OUTPUT_DC: '',
                amp.key_map.RIGHT_OUTPUT_DC: '',
                amp.key_map.LEFT_THERMAL_ERROR: '',
                amp.key_map.RIGHT_THERMAL_ERROR: '',
                }


        self.command_dict = {
                amp.key_map.LEFT_INPUT_DC: {'name': 'LEFT_INPUT_DC', 'set': self.amp_setting.SetLeftInputDC, 'get': self.amp_setting.GetLeftInputDC},
                amp.key_map.RIGHT_INPUT_DC: {'name': 'RIGHT_INPUT_DC', 'set': self.amp_setting.SetRightInputDC, 'get': self.amp_setting.GetRightInputDC},
                amp.key_map.LEFT_CLIPPING: {'name': 'LEFT_CLIPPING', 'set': self.amp_setting.SetLeftClipping, 'get': self.amp_setting.GetLeftClipping},
                amp.key_map.RIGHT_CLIPPING: {'name': 'RIGHT_CLIPPING', 'set': self.amp_setting.SetRightClipping, 'get': self.amp_setting.GetRightClipping},
                amp.key_map.LEFT_OUTPUT_DC: {'name': 'LEFT_OUTPUT_DC', 'set': self.amp_setting.SetLeftOutputDC, 'get': self.amp_setting.GetLeftOutputDC},
                amp.key_map.RIGHT_OUTPUT_DC: {'name': 'RIGHT_OUTPUT_DC', 'set': self.amp_setting.SetRightOutputDC, 'get': self.amp_setting.GetRightOutputDC},
                amp.key_map.LEFT_THERMAL_ERROR: {'name': 'LEFT_THERMAL_ERROR', 'set': self.amp_setting.SetLeftThermalError, 'get': self.amp_setting.GetLeftThermalError},
                amp.key_map.RIGHT_THERMAL_ERROR: {'name': 'RIGHT_THERMAL_ERROR', 'set': self.amp_setting.SetRightThermalError, 'get': self.amp_setting.GetRightThermalError},
                amp.key_map.LEFT_NEG_15VOLT_ERROR: {'name': 'LEFT_NEG_15VOLT_ERROR', 'set': self.amp_setting.SetLeftNeg15VoltError, 'get': self.amp_setting.GetLeftNeg15VoltError},
                amp.key_map.RIGHT_NEG_15VOLT_ERROR: {'name': 'RIGHT_NEG_15VOLT_ERROR', 'set': self.amp_setting.SetRightNeg15VoltError, 'get': self.amp_setting.GetRightNeg15VoltError},
                amp.key_map.LEFT_POS_15VOLT_ERROR: {'name': 'LEFT_POS_15VOLT_ERROR', 'set': self.amp_setting.SetLeftPos15VoltError, 'get': self.amp_setting.GetLeftPos15VoltError},
                amp.key_map.RIGHT_POS_15VOLT_ERROR: {'name': 'RIGHT_POS_15VOLT_ERROR', 'set': self.amp_setting.SetRightPos15VoltError, 'get': self.amp_setting.GetRightPos15VoltError},
                amp.key_map.LEFT_NEG_IV_ERROR: {'name': 'LEFT_NEG_IV_ERROR', 'set': self.amp_setting.SetLeftNegIVError, 'get': self.amp_setting.GetLeftNegIVError},
                amp.key_map.RIGHT_NEG_IV_ERROR: {'name': 'RIGHT_NEG_IV_ERROR', 'set': self.amp_setting.SetRightNegIVError, 'get': self.amp_setting.GetRightNegIVError},
                amp.key_map.LEFT_POS_IV_ERROR: {'name': 'LEFT_POS_IV_ERROR', 'set': self.amp_setting.SetLeftPosIVError, 'get': self.amp_setting.GetLeftPosIVError},
                amp.key_map.RIGHT_POS_IV_ERROR: {'name': 'RIGHT_POS_IV_ERROR', 'set': self.amp_setting.SetRightPosIVError, 'get': self.amp_setting.GetRightPosIVError},
                amp.key_map.TRIG_12V: {'name': 'TRIG_12V', 'set': self.amp_setting.SetTrig12V, 'get': self.amp_setting.GetTrig12V},
                amp.key_map.POWER_ON: {'name': 'POWER_ON', 'set': self.amp_setting.SetPowerOn, 'get': self.amp_setting.GetPowerOn},
                amp.key_map.DIAG1: {'name': 'MUTE1HZ', 'set': self.amp_setting.SetMute1Hz, 'get': self.amp_setting.GetMute1Hz},
                amp.key_map.DIAG2: {'name': 'MUTE10HZ', 'set': self.amp_setting.SetMute10Hz, 'get': self.amp_setting.GetMute10Hz},
                amp.key_map.DIAG3: {'name': 'Diagnostic', 'set': self.amp_setting.SetDiagnostic, 'get': self.amp_setting.GetDiagnostic},
                'mains_error': {'name': 'LEFT_RIGHT_MAINS_VOLTAGE_ERROR', 'set': self.amp_setting.SetMainsError, 'get': self.amp_setting.GetMainsError}

                }

        self.setting_gpios = {
                "L2100+1": {"gpio_pin": 9, "state": False},
                "L2100+2": {"gpio_pin": 87, "state": False},
                "R2100+1": {"gpio_pin": 77, "state": False},
                "R2100+2": {"gpio_pin": 78, "state": False},
                "LMono": {"gpio_pin": 81, "state": False},
                "RMono": {"gpio_pin": 76, "state": False},
                "Jumper1": {"gpio_pin": 13, "state": False}
            }

        self.mains_11_21_limits = {
                "On_Over_Critical": 0.815,
                "On_Over_Warning": 0.789,
                "On_Under_Critical": 0.543,
                "Off_Over_Critical": 0.803,
                "Off_Under_Critical": 0.556
                }

        #TODO These limits tbd, will get values from Boulder Amplifiers
        self.mains_3x_limits = {
                "On_Over_Critical": 0.815,
                "On_Over_Warning": 0.789,
                "On_Under_Critical": 0.543,
                "Off_Over_Critical": 0.803,
                "Off_Under_Critical": 0.556
                }
        self.mains_limits = {}
        self.mute_channel('all', True)
        self.load_fatal_errors()
####################
# Setup Functions  #
####################

    @LockCFunction
    def find_id(self):
        self.mac = self.amp_setting.GetMAC()
        self.room_table[self.mac] = 'self'
        devices = amp.DeviceVector()
        self.amp_setting.GetRoomDevices(devices)
        for i in range(len(devices)):
            self.room_table[devices[i].deviceMAC] = devices[i].deviceName
        MAC_list = self.room_table.keys()
        MAC_list.sort()
        self.id = MAC_list.index(self.mac)
        self.power_time_1 = self.id * 11 * 0.262

    #Find better way
    def add_key_handle(self, key_handle):
        self.key_handle = key_handle

    @LockCFunction
    def init_settings(self):
        bllogger('debug', 'init_settings')
        if 'arm' in platform.machine():
            for setting in  self.setting_gpios:
                fname = os.path.join("/sys","class","gpio","gpio%d"%self.setting_gpios[setting]["gpio_pin"],"value")
                with open(fname, 'r') as fp:
                    line = fp.readline()
                    self.setting_gpios[setting]['state'] = bool(int(line))
            if self.setting_gpios["Jumper1"]["state"]:
                    time.sleep(5)
                    self.power_on()
             
            if self.setting_gpios["L2100+1"]["state"] and self.setting_gpios["L2100+2"]["state"] and self.setting_gpios["R2100+1"]["state"] and self.setting_gpios["R2100+2"]["state"]:
                version_num = 3000
                self.mains_limits = self.mains_3x_limits

            elif not self.setting_gpios["L2100+1"]["state"] and not self.setting_gpios["L2100+2"]["state"] and not self.setting_gpios["R2100+1"]["state"] and not self.setting_gpios["R2100+2"]["state"]:
                version_num = 2100
                self.mains_limits = self.mains_11_21_limits

            elif not self.setting_gpios["L2100+1"]["state"] and self.setting_gpios["L2100+2"]["state"] and not self.setting_gpios["R2100+1"]["state"] and self.setting_gpios["R2100+2"]["state"]:
                version_num = 1100
                self.mains_limits = self.mains_11_21_limits
            else:
                bllogger('err', "Version pins not in reconized configureation")
                self.mains_limits = self.mains_11_21_limits
                version_num = 0000
            
            if self.setting_gpios['LMono']['state']:
                version_num += 60
            else:
                version_num += 50

            self.amp_setting.SetModelVersion(version_num)
            
            self.amp_setting.SetLeftMono(not self.setting_gpios["LMono"]["state"])
            self.amp_setting.SetRightMono(not self.setting_gpios["RMono"]["state"])
        else:
            config = amp.GetSimpleConfig()
            model_version = config.GetAsInt("model_version",2100)
            self.amp_setting.SetModelVersion(model_version)
            print model_version/1000
            if model_version/1000 == 1 or model_version/1000 == 2:
                self.mains_limits = self.mains_11_21_limits
            elif model_version/1000 == 3:
                self.mains_limits = self.mains_3x_limits

            right_mono = config.GetAsBool("right_mono",False)
            left_mono = config.GetAsBool("left_mono", False)

            self.amp_setting.SetRightMono(right_mono)
            self.amp_setting.SetLeftMono(left_mono)

#####################################

#################
#  Errors       #
#################


    def add_error(self, error, state):
        with self.lock:
            self.error_dict[error] = state

    def clear_errors(self):
        with self.lock:
            self.error_dict.clear()

    def store_errors(self):
        for error in self.error_dict:
            if self.command_dict[error]['get']() != self.error_dict[error]:
                if error == amp.key_map.POWER_ON:
                    continue
                else:
                        self.command_dict[error]['set'](self.error_dict[error])

            if self.amp_setting.GetPowerOn():
                if error in self.transient_errors:
                    self.transient_errors[error] = str(datetime.datetime.now())

    def check_fatal(self):
	errors = self.get_fatal_errors()
	if len(errors):
	    return True
	else:
	    return False

    def get_fatal_errors(self):
        error_list = []
        for error in self.fatal_errors:
            value = self.command_dict[error]['get']()
            if value:
                error_list.append(self.command_dict[error]['name'])
        return error_list

    def load_fatal_errors(self):
        for fatal in self.fatal_errors:
            key = self.command_dict[fatal]['name']
            error = self.config.Get(str(key), "")
	    if error is not "":
                self.command_dict[fatal]['set'](True)


    def get_fatal_error_history(self):
	set_errors = {}
        for fatal in self.fatal_errors:
            key = self.command_dict[fatal]['name']
            #key = fatal
            error = self.config.Get(str(key), "")
	    if error is not "":
		set_errors[str(key)] = error
	    #bllogger('debug', "%r : %r" % (fatal, error))
	return set_errors


    def fatal(self, error):
	bllogger('err', 'Fatal Error: %s' % error)
        self.config.Set(self.command_dict[error]['name'], str(datetime.datetime.now()), True)
        self.store_errors();
        self.abort()

    @LockCFunction
    def clear_fatals(self):
        for error in self.fatal_errors:
            self.config.Set(self.command_dict[error]['name'],'', True)
            self.command_dict[error]['set'](False)
        if self.GetPowerOn():
            self.led_state.set_state('on')
        else:
            self.led_state.set_state('standby')

    def are_errors(self,state):
        if state == 'left':
            error_list = filter(lambda x: 'LEFT' in self.command_dict[x]['name'] , self.transient_errors)
        elif state == 'right':
            error_list  = filter(lambda x: 'RIGHT' in self.command_dict[x]['name'] , self.transient_errors )
        elif state == 'all':
            error_list = self.transient_errors
        pruned_list = error_list
        for error in pruned_list:
            if error is not amp.key_map.RIGHT_CLIPPING and error is not amp.key_map.LEFT_CLIPPING:
                if self.command_dict[error]['get']():
                    return True
        if self.command_dict['mains_error']['get']() != amp.mains_state.Clear:
            return True
        return False

    def get_trans_errors(self):
        error_list = []
        for error in self.transient_errors:
            value = self.command_dict[error]['get']()
            if value:
                error_list.append(self.command_dict[error]['name'])
        return error_list

    def get_trans_history(self):
        error_dict = {}
        for error in self.transient_errors:
            time_stamp = self.transient_errors[error] 
            if time_stamp != '':
                error_dict[self.command_dict[error]['name']] =  time_stamp 
        return error_dict

    def power_on_error_check(self):
        #this is for regestering errors that are asserted during powerup and are still present
        for error in self.transient_errors:
            if self.command_dict[error]['get']():
                self.transient_errors[error] = str(datetime.datetime.now())
        for error in self.fatal_errors:
            if self.command_dict[error]['get']():
                self.config.Set(self.command_dict[error]['name'], str(datetime.datetime.now()), True)
        pass


########################

######################
#  Power Control     #
######################
    def GetPowerOn(self):
        return self.amp_setting.GetPowerOn()

    def wakeOnLan(self):
	wakeup = self.amp_setting.GetWakeOnLan()
	if wakeup:
	    self.amp_setting.SetWakeOnLan(False)
	return wakeup

    
    def relay_wait(self, relay_time):
        t1 = time.time()
        with self.lock:
            while (time.time()-t1 < relay_time):
                self.key_handle.process()
                if amp.key_map.POWER_ON in self.error_dict:
                    if self.error_dict[amp.key_map.POWER_ON]:
                        self.abort()
                        return False
                if amp.key_map.TRIG_12V in self.error_dict:
                    if not self.error_dict[amp.key_map.TRIG_12V]:
                        self.abort()
                        return False
                self.store_errors()
                self.clear_errors()
        return True


    def abort(self):
        self.amp_setting.SetSurge(False)
        self.power_off()


    def power_off(self):
        bllogger('debug', 'power_off')
        self.gpio_handle.relays_off()
        self.mute_channel('all', True)
        self.amp_setting.SetPowerOn(False)
        self.led_state.set_state('standby')

    
    def power_on(self):
        bllogger('debug', 'power_on')
        if self.amp_setting.update_failed_:
            bllogger('debug', "Resetting failed update state")
            self.led_state.set_state('standby')
            self.amp_setting.update_failed_ = False
            self.amp_setting.usb_update_ready_ = False
            return
        if self.amp_setting.GetUpdateInProgress():
            bllogger('debug', "Abort power on, updating the amplifier")
            return
        if self.get_usb_update_available():
            self.start_update()
            return
        if self.amp_setting.GetPowerOn():
            bllogger('debug', "Abort power on, amplifier already on")
            return
        if self.fatal_shutdown:
            bllogger('debug', "Abort power on, in a fatal error shutdown")
            return
        if self.amp_setting.GetMainsError() != amp.mains_state.Clear:
            bllogger('debug', "Abort power on, mains voltage in bad state")
            return

        # Remove the event that causes power on so that we don't double count it in relay_wait
        if amp.key_map.POWER_ON in self.error_dict:
            self.error_dict.pop(amp.key_map.POWER_ON)
        if amp.key_map.TRIG_12V in self.error_dict:
            self.error_dict.pop(amp.key_map.TRIG_12V)

        self.led_state.set_state('turning_on')
        self.amp_setting.SetSurge(True)
        if not self.relay_wait(self.power_time_1):
            return False
        self.gpio_handle.set(amp.gpio_map.RELAY_A, True)
        self.gpio_handle.set(amp.gpio_map.RELAY_C, True)

        if not self.relay_wait(self.power_time_2):
            return False
        self.gpio_handle.set(amp.gpio_map.RELAY_E, True)
        self.gpio_handle.set(amp.gpio_map.RELAY_H, True)

        if not self.relay_wait(self.surge_time_1):
            return False
        self.gpio_handle.set(amp.gpio_map.RELAY_B, True)
        self.gpio_handle.set(amp.gpio_map.RELAY_D, True)
        time.sleep(0.04)
        self.gpio_handle.set(amp.gpio_map.RELAY_A, False)
        self.gpio_handle.set(amp.gpio_map.RELAY_C, False)

        if not self.relay_wait(self.surge_time_2):
            return False
        self.gpio_handle.set(amp.gpio_map.RELAY_F, True)
        self.gpio_handle.set(amp.gpio_map.RELAY_G, True)
        time.sleep(0.04)
        self.gpio_handle.set(amp.gpio_map.RELAY_E, False)
        self.gpio_handle.set(amp.gpio_map.RELAY_H, False)

        self.relay_wait(self.safe_time)
        self.amp_setting.SetSurge(False)
        self.amp_setting.SetPowerOn(True)
        self.power_on_error_check()
        self.led_state.set_state('on')
        return True

######################

    def thermal_wait(self, wait_time):
        t1 = time.time()
        while (time.time()-t1 < wait_time):
            self.key_handle.thermal_wait()

    def mute_channel(self, channel, state):
        #don't unmute channels while the power is off 
        if not self.amp_setting.GetPowerOn() and not state:
            return
        if  self.amp_setting.GetMute1Hz() or  self.amp_setting.GetMute10Hz():
            return
        if channel.lower() == 'all':
            self.gpio_handle.all_mute(state)
	    if self.amp_setting.GetLeftMute() != state:
		    self.amp_setting.SetLeftMute(state)
	    if self.amp_setting.GetRightMute() != state:
		    self.amp_setting.SetRightMute(state)
        elif channel.lower() == 'left':
            self.gpio_handle.left_mute(state)
	    if self.amp_setting.GetLeftMute() != state:
		self.amp_setting.SetLeftMute(state)
        elif channel.lower() == 'right':
            self.gpio_handle.right_mute(state)
	    if self.amp_setting.GetRightMute() != state:
		self.amp_setting.SetRightMute(state)

    def get_mains_value(self):
        if 'arm' in platform.machine():
            mains_file = '/sys/bus/iio/devices/iio:device0/in_voltage3_raw'
        else:
            mains_file = 'mains_voltage'
        try:
            with open(mains_file, 'r') as fp:
                value = fp.readline().strip()
        except:
            bllogger('err', 'Could not read mains voltage from %s' % mains_file)
            value  = -1

        return value

    def check_mains_limits(self):
        mains_voltage = (float(self.get_mains_value())/4096)*1.8
        if  self.amp_setting.GetPowerOn():
            if mains_voltage >= self.mains_limits["On_Over_Critical"]:
                self.set_mains_state('critical')                
                self.command_dict['mains_error']['set'](amp.mains_state.Over_Critical)

            elif mains_voltage >= self.mains_limits["On_Over_Warning"]:
                self.set_mains_state('warning')                
                self.command_dict['mains_error']['set'](amp.mains_state.Over_Warning)

            elif mains_voltage <= self.mains_limits["On_Under_Critical"] and mains_voltage > 0:
                self.set_mains_state('critical')                
                self.command_dict['mains_error']['set'](amp.mains_state.Under_Critical)

            else:
                self.set_mains_state('clear')
                self.command_dict['mains_error']['set'](amp.mains_state.Clear)
        else:
            if mains_voltage >= self.mains_limits["Off_Over_Critical"]:
                self.set_mains_state('critical')                
                self.command_dict['mains_error']['set'](amp.mains_state.Over_Critical)

            elif mains_voltage <= self.mains_limits["Off_Under_Critical"] and mains_voltage > 0:
                self.set_mains_state('critical')                
                self.command_dict['mains_error']['set'](amp.mains_state.Under_Critical)

            else:
                self.set_mains_state('clear')
                self.command_dict['mains_error']['set'](amp.mains_state.Clear)

    def set_mains_state(self, state):
        if state == 'critical':
            self.mute_channel('all', True)
            self.abort()
            self.led_state.set_state('mains_critical')
        elif state == 'warning':
            self.mute_channel('all', True)
            self.led_state.set_state('mains_warning')
        elif state == 'clear':
            pass
        else:
            bllogger('err', "Unknown state of mains voltage")

                
    def set_room(self, room):
        bllogger('info', 'AmplifierSettings.py set_room')
        self.room = room
        self.amp_setting.SetRoom(str(room))

    def get_update_available(self):
        self.update_available = self.amp_setting.CheckForUpdates()
        return self.update_available

    def get_usb_update_available(self):
        return self.amp_setting.usb_update_ready_;

    def start_update(self):
        #want the amplifier off while updating
        self.abort()
        if self.get_update_available():
            self.amp_setting.InstallUpdate()
            self.led_state.set_state('updating')
            return True
        else:
            return False

    def get_model_name(self):
        return self.amp_setting.GetLeftVersion()

    def process_state(self):
        config = amp.GetConfigLock() 
        amplifier = amp.GetAmplifierSettingLock()
        fatal_error_w_diag = False
        self.find_id()
        devices = amp.DeviceVector()
        self.amp_setting.GetRoomDevices(devices)
        for i  in range(len(devices)):
            bllogger('info', 'Share room with %r: %r' % (devices[i].deviceName, devices[i].deviceMAC))

        if len(self.error_dict): 
            bllogger('info', 'Processing state, with errors %r' % self.error_dict)

        #power button pressed 
        if self.amp_setting.upnp_poweroff_flag_ :
            self.abort()
            self.amp_setting.upnp_poweroff_flag_ = False
        else:
            if amp.key_map.POWER_ON in self.error_dict and self.error_dict[amp.key_map.POWER_ON]:
                power_state = not self.command_dict[amp.key_map.POWER_ON]['get']()
                #self.command_dict[amp.key_map.POWER_ON]['set'](power_state)
                if power_state: 
                    amp.ReleaseAmplifierSettingLock(amplifier)
                    amp.ReleaseConfigLock(config)
                    self.power_on()
                    config = amp.GetConfigLock() 
                    amplifier = amp.GetAmplifierSettingLock()
                else:
                    self.abort()
	if self.wakeOnLan():
            power_on =  self.command_dict[amp.key_map.POWER_ON]['get']()
	    bllogger('debug', 'python wakeonlan check, Power on: %d' % power_on)
	    if not power_on:
                amp.ReleaseAmplifierSettingLock(amplifier)
                amp.ReleaseConfigLock(config)
                self.power_on()
                config = amp.GetConfigLock() 
                amplifier = amp.GetAmplifierSettingLock()

        if amp.key_map.TRIG_12V in self.error_dict:
            trig = self.error_dict[amp.key_map.TRIG_12V]
            if trig:
                if not self.GetPowerOn():
                    amp.ReleaseAmplifierSettingLock(amplifier)
                    amp.ReleaseConfigLock(config)
                    self.power_on()
                    config = amp.GetConfigLock() 
                    amplifier = amp.GetAmplifierSettingLock()
            else:
                if self.GetPowerOn():
                    self.abort()

        if self.amp_setting.update_failed_:
            self.led_state.set_state('update_failure')
            amp.ReleaseAmplifierSettingLock(amplifier)
            amp.ReleaseConfigLock(config)
            return True
        
        if self.amp_setting.GetDiagnostic():
            if len(self.get_fatal_errors()):
                self.clear_fatals()
            #remove fatal errors from dict
            if len(self.error_dict):
                for fatal in self.fatal_errors:
                    rtn = self.error_dict.pop(fatal, None) 
                    #rtn is the value of the fatal error
                    if rtn:
                        fatal_error_w_diag = True
             
        self.check_mains_limits()
        if not  self.command_dict[amp.key_map.POWER_ON]['get']():
            self.store_errors()     # need to keep track of transient errors even if we're not changing mute/LEDs while power is off
            self.clear_errors()
            if not self.amp_setting.GetSurge():
                if self.led_state.queued("mains_critical") and self.led_state.queued('updating'):
                    self.led_state.set_state('standby')
            self.mute_state.set_state('off')
            amp.ReleaseAmplifierSettingLock(amplifier)
            amp.ReleaseConfigLock(config)
            return True

        #####  rest of function only gets run if the amplifier is on ####
        #Need to check .cpp variables if fatals set during power off
        for error in self.error_dict.keys():
            if error in self.fatal_errors:
                if self.error_dict[error]:
                    self.fatal(error)
                    self.clear_errors()
                    amp.ReleaseAmplifierSettingLock(amplifier)
                    amp.ReleaseConfigLock(config)
                    return False

        if self.amp_setting.GetMute1Hz():
            self.mute_state.set_state('1hz_mute')
        elif self.amp_setting.GetMute10Hz():
            self.mute_state.set_state('10hz_mute')
        else:
            self.mute_state.set_state('off')


        if self.get_fatal_errors():
            self.abort()
            self.clear_errors()
            amp.ReleaseAmplifierSettingLock(amplifier)
            amp.ReleaseConfigLock(config)
            return False


        #For a Mono amplifier
        if self.amp_setting.GetLeftMono():
            #bllogger('debug', 'Mono Amp')

            #ignore OutputDC when in thermal shutdown
            if self.command_dict[amp.key_map.LEFT_THERMAL_ERROR]['get']():
                if amp.key_map.LEFT_OUTPUT_DC in self.error_dict:
                    self.error_dict.pop(amp.key_map.LEFT_OUTPUT_DC)

            self.store_errors()
            #are there any errors
            if self.are_errors('all'):
                if not self.amp_setting.GetDiagnostic():
                    self.mute_channel('all', True)
            else:
                self.mute_channel('all', False)
                
        else:
            #bllogger('debug', 'Stereo Amp')

            if self.command_dict[amp.key_map.LEFT_THERMAL_ERROR]['get']():
                if amp.key_map.LEFT_OUTPUT_DC in self.error_dict:
                    self.error_dict.pop(amp.key_map.LEFT_OUTPUT_DC)
            
            if self.command_dict[amp.key_map.RIGHT_THERMAL_ERROR]['get']():
                if amp.key_map.RIGHT_OUTPUT_DC in self.error_dict:
                    self.error_dict.pop(amp.key_map.RIGHT_OUTPUT_DC)

            self.store_errors()
            if self.are_errors('left'):
                if not self.amp_setting.GetDiagnostic():
                    self.mute_channel('left', True)
            else:
                self.mute_channel('left', False)

            if self.are_errors('right'):
                if not self.amp_setting.GetDiagnostic():
                    self.mute_channel('right', True)
            else:
                self.mute_channel('right', False)
        
        if len(self.get_trans_errors()) or fatal_error_w_diag:
            self.led_state.set_state('transient_error')
        else:
            power_on =  self.command_dict[amp.key_map.POWER_ON]['get']()
            if power_on:
                self.led_state.set_state('on')
            else:
                self.led_state.set_state('standby')
                

        self.clear_errors()
        amp.ReleaseAmplifierSettingLock(amplifier)
        amp.ReleaseConfigLock(config)
        return True

        
